/************************************************************************

    TIMEX2P3 1.0.0 - Convert Timex FDD3000 disc images to +3DOS

    Copyright (C) 2007, 2008  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "config.h"
#include "libdsk.h"
#include "plus3fs.h"

#ifdef __PACIFIC__
#define AV0 "TIMEX2P3"
#else
#define AV0 argv[0]
#endif

static const char *gl_itype = NULL;
static const char *gl_otype = "dsk";
static const char *gl_ocomp = NULL;
static const char *gl_icomp = NULL;
static int gl_cpm = 0;
static int gl_xlt = 0;
static int gl_format = 0;
static int gl_dos = 0;

static int gl_secshift = 4;
static int gl_dirlen = 128;

static unsigned char timex_dir[4096];
static int gl_skew[16];
PLUS3FS gl_fs;
DSK_PDRIVER gl_timex;

static DSK_GEOMETRY timex_geom =
{
	SIDES_ALT,
	40,	/* 40 cyls */
	1,	/* 1 head */
	16,	/* 16 sectors */
	0,	/* starting at 0 */
	256,	/* 256-byte sectors */
	RATE_SD,/* Data rate */
	12,	/* R/W gap */
	23,	/* Format gap */
	0,	/* FM */
	1,	/* Disable multitrack */
	1	/* Skip deleted data */
};

static dsk_err_t load_timex_track(unsigned char *buf, int track)
{
	int n;
	dsk_err_t err;

	for (n = 0; n < 16; n++)
	{
		err = dsk_pread(gl_timex, &timex_geom,
				buf + 256 * n, track, 0, gl_skew[n]);
		if (err) return err;
	}
/*
   {FILE *fp;
	fp = fopen("tmp", "wb");
	fwrite(buf, 1, 4096, fp);
	fclose(fp);} */
	return DSK_ERR_OK;
}


static dsk_err_t load_timex_dir(void)
{
	return load_timex_track(timex_dir, gl_cpm ? 2 : 4);
}

typedef struct
{
	unsigned char dirfcb[32];
	unsigned long size;
	unsigned long pos;
	unsigned localpos;
	unsigned char extent[16384];
} TIMEXFILE;

dsk_err_t timex_loadblock(unsigned char *buf, unsigned block)
{
	dsk_lsect_t lsect;
	dsk_phead_t head;
	dsk_pcyl_t cyl;
	dsk_psect_t sec;
      	dsk_err_t err;
	int n;

	/* Logical sector number: (4 * block number) + no. of system sectors */
	lsect = (block * gl_secshift) + (gl_cpm ? 32 : 64);
	for (n = 0; n < gl_secshift; n++, lsect++)
	{
		dg_ls2ps(&timex_geom, lsect, &cyl, &head, &sec);
		err = dsk_pread(gl_timex, &timex_geom, buf, cyl, 0, gl_skew[sec]);
		if (err) return err;
		buf += timex_geom.dg_secsize;
	}
	return DSK_ERR_OK;
}

dsk_err_t timex_loadextent(TIMEXFILE *tfp)
{
	int n;
	dsk_err_t err;

	memset(tfp->extent, 0, sizeof(tfp->extent));
	for (n = 0; n < 16; n++)
	{
		if (tfp->dirfcb[n + 16])
		{
			err = timex_loadblock(tfp->extent + 1024 * n,
				tfp->dirfcb[n + 16]);
			if (err) return err;	
		}
	}
	return DSK_ERR_OK;
}


dsk_err_t timex_open(TIMEXFILE *tfp, unsigned char *dirent)
{
	int n;

	memcpy(tfp->dirfcb, dirent, 32);
	tfp->size = 0;
	tfp->pos  = 0;
	tfp->localpos = 0;
	/* Now find all the extents */
	for (n = 0; n < gl_dirlen; n++)
	{
		dirent = timex_dir + 32 * n;
		if (!memcmp(tfp->dirfcb, dirent, 12))	/* Same user & name */
		{
			tfp->size += 128 * dirent[15];
		}
	}
	/* The number of bytes used in the last record */
	n = tfp->dirfcb[13] & 0x7F;
	if (n) { tfp->size = (tfp->size - 128) + n; }	
/*
	printf("%-8.8s.%-3.3s length %ld\n", tfp->dirfcb + 1, tfp->dirfcb+9,
			tfp->size); */
	return timex_loadextent(tfp);
}

dsk_err_t timex_close(TIMEXFILE *tfp)
{
	/* does nothing for now */
	return DSK_ERR_OK;
}

dsk_err_t timex_getc(TIMEXFILE *tfp, int *c)
{
	dsk_err_t err;
	int n;
	unsigned char *dirent;

	if (tfp->pos >= tfp->size)
	{
		*c = EOF;
		return DSK_ERR_OK;
	}

	if (tfp->localpos >= sizeof(tfp->extent))
	{
		int found = 0;
		/* Find the next extent */
/*		printf("Seeking %-8.8s.%-3.3s  ext=%d\n",
				tfp->dirfcb + 1, tfp->dirfcb+9, 
				tfp->dirfcb[12]); */
		for (n = 0; n < gl_dirlen; n++)
		{
			dirent = timex_dir + 32 * n;
			/* Same user & name, next extent */
			if (dirent[12] == (tfp->dirfcb[12] + 1) &&
			   !memcmp(tfp->dirfcb, dirent, 12))
			{
				memcpy(tfp->dirfcb, dirent, 32);
				found = 1;
				break;
			}
		}
		if (!found) { *c = EOF; return DSK_ERR_OK; }
		err = timex_loadextent(tfp);
		if (err) return err;
		tfp->localpos = 0;
	}	
	*c = tfp->extent[tfp->localpos];
	++tfp->localpos;
	++tfp->pos;
	return DSK_ERR_OK;	
}




static dsk_err_t convert_file(unsigned char *dirent)
{
	TIMEXFILE tfp;
	PLUS3FILE p3fp;
	dsk_err_t err;
	int n, c, pos;
	int hdrlen;
	unsigned char header[7];
	unsigned char namebuf[11];
	unsigned char p3dos[128];
	unsigned long filelen;

	err = timex_open(&tfp, dirent);
	if (err) return err;
/* Create a +3 file */

	memcpy(namebuf, dirent + 1, 11);
	for (c = 0; c < 11; c++)
	{
		if (namebuf[c] == '*' || namebuf[c] == '?')
		{
			namebuf[c] = '_';
		}
		if (islower(namebuf[c] & 0x7F)) 
		{
			namebuf[c] = (namebuf[c] & 0x80) | toupper(namebuf[c]);
		}
	}
	err = p3fs_creat(gl_fs, &p3fp, dirent[0], (char *)namebuf);
	if (err) return err;

	if (gl_xlt)	/* do wacky stuff */
	{
		hdrlen = 7;
		for (pos = 0; pos < hdrlen; pos++)
		{
			err = timex_getc(&tfp, &c);
			if (err) return err;
			header[pos] = c;
			if (c == EOF) break;
			if (pos == 0) switch(c)
			{
				case 0: hdrlen = 7; break;
				case 1: hdrlen = 5; break;
				case 2: hdrlen = 5; break;
				case 3: hdrlen = 5; break;
			}
		}
		if (pos < hdrlen) /* Not all of the header was loaded. 
				   * Regurgitate it and call it a day. */
		{
			for (n = 0; n < pos; n++)
			{
				err = p3fs_putc(header[n], p3fp);
				if (err) return err;
			}
			timex_close(&tfp);
			return p3fs_close(p3fp);
		}
		/* OK. Header loaded. */
		if (header[0] >= 4)	/* Unrecognised filetype. */
		{			/* Regurgitate it. */
			for (n = 0; n < 7; n++)
			{
				err = p3fs_putc(header[n], p3fp);
				if (err) return err;
			}
		}
		else	/* Convert to +3DOS header */
		{
			filelen = (tfp.size - hdrlen) + 128;
			memset(p3dos, 0, sizeof(p3dos));
			memcpy(p3dos, "PLUS3DOS\032\001", 10);
			p3dos[11] = (filelen      ) & 0xFF;
			p3dos[12] = (filelen >>  8) & 0xFF;
			p3dos[13] = (filelen >> 16) & 0xFF;
			p3dos[14] = (filelen >> 24) & 0xFF;
			p3dos[15] = header[0];
			if (header[0] == 0)	/* BASIC */
			{
/* Not entirely sure about the order of the header fields here */
				p3dos[16] = header[3]; 
				p3dos[17] = header[4];
				p3dos[18] = header[1];
				p3dos[19] = header[2];
				p3dos[20] = header[5];
				p3dos[21] = header[6];
			}
			else
			{
				p3dos[16] = header[1];
				p3dos[17] = header[2];
				p3dos[18] = header[3];
				p3dos[19] = header[4];
			}

/* Checksum the generated +3DOS header */	
			for (n = 0; n < 127; n++)
			{
				p3dos[127] += p3dos[n];
			}
/* And write it out */
			for (n = 0; n < 128; n++)
			{
				err = p3fs_putc(p3dos[n], p3fp);
				if (err) return err;
			}
		}
	}
	do
	{
		err = timex_getc(&tfp, &c);
		if (err) return err;
		if (c == EOF) break;
		err = p3fs_putc(c, p3fp);
		if (err) return err;
	} while (c != EOF);

	timex_close(&tfp);
	return p3fs_close(p3fp);
}


static dsk_err_t convert_all(void)
{
	int n;
	unsigned char *dirent;
	dsk_err_t err;

	for (n = 0; n < gl_dirlen; n++)
	{
		dirent = timex_dir + 32 * n;
		if (dirent[0] == 0xE5) continue; /* Deleted or unused */
		if (dirent[0] == 0xFF && dirent[16] == 0)
		{
			err = p3fs_setlabel(gl_fs, (char *)(dirent + 1));
			if (err) return err;
			continue;
		}
		if (dirent[0] == 0xFF)
		{
			fprintf(stderr, "Skipped label: <%d>%-8.8s.%-3.3s\n",
				dirent[16], dirent + 1, dirent + 9);
			continue;
		}
		/* First extent of a file? */
		if (dirent[0] < 16 && dirent[12] == 0)
		{
			err = convert_file(dirent);
			if (err)
			{
				fprintf(stderr, "Failed to convert file: "
					"<%d>%-8.8s.%-3.3s (%s)\n",
					dirent[0], dirent + 1, dirent + 9,
				p3fs_strerror(err));
			}
		}
	}
	return DSK_ERR_OK;
}



static void report(const char *s)
{
        printf("%s\r", s);
        fflush(stdout);
}

static void report_end(void) 
{
        printf("\r%-79.79s\r", "");
        fflush(stdout);
}

static unsigned char boot_default[] =
{
        0x00, /* Disc type */
        0x00, /* Disc geometry */
        0x28, /* Tracks */
        0x09, /* Sectors */
        0x02, /* Sector size */
        0x01, /* Reserved tracks */
        0x03, /* Block shift */
        0x06, /* Directory blocks */
        0x2A, /* Gap length (R/W) */
        0x52  /* Gap length (format) */
};


static unsigned char boot_180[] =
{
        0x00, /* Disc type */
        0x00, /* Disc geometry */
        0x28, /* Tracks */
        0x09, /* Sectors */
        0x02, /* Sector size */
        0x01, /* Reserved tracks */
        0x03, /* Block shift */
        0x02, /* Directory blocks */
        0x2A, /* Gap length (R/W) */
        0x52  /* Gap length (format) */
};

static unsigned char boot_720[] =
{
        0x03, /* Disc type */
        0x81, /* Disc geometry */
        0x50, /* Tracks */
        0x09, /* Sectors */
        0x02, /* Sector size */
        0x02, /* Reserved tracks */
        0x04, /* Block shift */
        0x04, /* Directory blocks */
        0x2A, /* Gap length (R/W) */
        0x52  /* Gap length (format) */
};
void syntax(const char *arg)
{
	fprintf(stderr, "Syntax: %s {options} infile.dsk outfile.dsk\n\n", arg);
	fprintf(stderr, "Options:\n"
			"-itype type: Input file type (default=autodetect)\n"
			"-otype type: Output file type (default=dsk)\n"
			"-icomp type: Input compression (default=autodetect)\n"
			"-ocomp type: Output compression (default=none)\n"
			"-cpm:        Input disk is CP/M format\n"
			"-xlt:        Attempt to convert file headers\n"
			"-180:        Output 180k disc image\n"
			"-720:        Output 720k disc image\n"
			"-dos:        Output DOS disc image\n");

	exit(1);
}

int main(int argc, char **argv)
{
	char *inname = NULL;
	char *outname = NULL;
	int n, optend = 0;
	const char *option, *value;
	dsk_err_t err;

	for (n = 1; n < argc; n++)
	{
		if (!strcmp(argv[n], "--"))
		{
			optend = 1;
			continue;
		}
		if (optend || argv[n][0] != '-')
		{
			if      (inname  == NULL) inname = argv[n];
			else if (outname == NULL) outname = argv[n];
			else    syntax(AV0);
			continue;
		}
		option = argv[n] + 1;
		if (option[0] == '-') ++option;

		if      (!strcmp (option, "180")) 
		{
			gl_format = 180;
			continue;
		}
		else if (!strcmp (option, "720"))
		{
			gl_format = 720;
			continue;
		}
		else if (!strcmp (option, "dos")) 
		{
			gl_dos = 1;	
			gl_format = 720;
			continue;
		}
		else if (!strcmp (option, "cpm")) 
		{
			gl_cpm = 1;	
			continue;
		}
		else if (!strcmp (option, "xlt")) 
		{
			gl_xlt = 1;	
			continue;
		}

/* Check for --variable=value and -variable value */
		value = strchr(option, '=');
		if (value) ++value;
		else
		{
			++n;
			if (n >= argc) syntax(AV0);
			value = argv[n];
		}
		if      (!strncmp(option, "itype", 5)) gl_itype = value;
		else if (!strncmp(option, "otype", 5)) gl_otype = value;
		else if (!strncmp(option, "icomp", 5)) gl_icomp = value;
		else if (!strncmp(option, "ocomp", 5)) gl_ocomp = value;
		else syntax(AV0);
	}
	if (inname == NULL || outname == NULL)
	{
		syntax(AV0);
	}
        dsk_reportfunc_set(report, report_end);
	for (n = 0; n < 16; n++)
	{
		if (gl_cpm) gl_skew[n] = (n * 5) % 16;
		else	    gl_skew[n] = (n * 7) % 16;
	}

	err = dsk_open(&gl_timex, inname, gl_itype, gl_icomp);
	if (!err) err = load_timex_dir();
	if (err)
	{
		fprintf(stderr, "Can't open %s: %s\n",
			inname, dsk_strerror(err));
		exit(1);
	}

	switch(gl_format)
	{
		case 180: err = p3fs_mkfs(&gl_fs, outname, gl_otype, gl_ocomp,
					boot_180, 0); break;
		case 720: err = p3fs_mkfs(&gl_fs, outname, gl_otype, gl_ocomp,
					boot_720, 1); break;
		default:  err = p3fs_mkfs(&gl_fs, outname, gl_otype, gl_ocomp,
					boot_default, 1); break;
	}

	if (err)
	{
		dsk_close(&gl_timex);
		fprintf(stderr, "Can't create %s: %s\n",
			outname, p3fs_strerror(err));
		exit(1);
	}
	err = convert_all();
	if (err)
	{
		dsk_strerror(err);
	}
	else if (gl_dos && (gl_format == 180 || gl_format == 720))
	{	
		err = p3fs_dossync(gl_fs, gl_format, gl_dos);
		if (err)
		{
			fprintf(stderr, "Cannot do DOS sync on %s: %s\n",
					outname, p3fs_strerror(err));
		
		}
	}
	dsk_close(&gl_timex);
	err = p3fs_umount(gl_fs);
	if (err) 
	{
		fprintf(stderr, "Cannot close %s: %s\n",
				outname, p3fs_strerror(err));
		return 1;
	}
	return 0;
}
